#[macro_use]
extern crate miette;

#[macro_use]
extern crate serde;

mod install;

use std::path::{Path, PathBuf};

use miette::{IntoDiagnostic, Result, WrapErr};

use install::{Pipeline, RemoteResource, ShortcutTarget, Step};

const SWTOOLS_PATH: &'static str = "C:\\SWTools";

pub fn swtools_path() -> &'static Path {
    Path::new(SWTOOLS_PATH)
}

pub fn desktop_path() -> PathBuf {
    Path::new("H:\\Profile\\Desktop").to_owned()
}

pub struct Context {
    pub reqwest: reqwest::Client,
}

#[tokio::main(flavor = "current_thread")]
async fn main() -> Result<()> {
    println!("Bootstrapping...");

    if cfg!(not(windows)) {
        bail!("Your platform is not supported.");
    }

    if !swtools_path().exists() {
        bail!("Could not find or access {}", SWTOOLS_PATH);
    }

    let nppp_zip = swtools_path().join("temp").join("notepad-plus-plus.zip");
    let nppp_dir = swtools_path().join("notepad-plus-plus");

    let epp_zip = swtools_path().join("temp").join("explorer-plus-plus.zip");
    let epp_dir = swtools_path().join("explorer-plus-plus");

    let minecraft_dir = swtools_path().join("minecraft");

    let psiphon_dir = swtools_path().join("psiphon");
    let psiphon_bin = psiphon_dir.join("psiphon3.exe");

    let utilities = [
        Pipeline::new(
            "Install Notepad++",
            vec![
                Step::DownloadFile {
                    file: nppp_zip.as_path().into(),
                    res: RemoteResource::GitHubArtifact {
                        repo: "notepad-plus-plus/notepad-plus-plus",
                        pattern: "npp.{{tag_name_strip_prefix}}.portable.x64.zip",
                    },
                },
                Step::ExtractFile {
                    file: nppp_zip.as_path().into(),
                    dest: nppp_dir.as_path().into(),
                },
                Step::CreateShortcut {
                    target: ShortcutTarget::Executable {
                        file: nppp_dir.join("notepad++.exe").into(),

                        args: "",
                    },
                    file: desktop_path().join("Notepad++.lnk").into(),
                },
            ],
        ),
        Pipeline::new(
            "Install Explorer++",
            vec![
                Step::DownloadFile {
                    file: epp_zip.as_path().into(),
                    res: RemoteResource::Url(
                        "https://explorerplusplus.com/software/explorer++_1.3.5_x64.zip",
                    ),
                },
                Step::ExtractFile {
                    file: epp_zip.as_path().into(),
                    dest: epp_dir.as_path().into(),
                },
                Step::CreateShortcut {
                    target: ShortcutTarget::Executable {
                        file: epp_dir.join("Explorer++.exe").into(),

                        args: "",
                    },
                    file: desktop_path().join("Explorer++.lnk").into(),
                },
            ],
        ),
        Pipeline::new(
            "Install Psiphon VPN",
            vec![
                Step::DownloadFile {
                    file: psiphon_bin.as_path().into(),
                    res: RemoteResource::Url(
                        "https://s3.amazonaws.com/f58p-mqce-k1yj/psiphon3.exe",
                    ),
                },
                Step::CreateShortcut {
                    target: ShortcutTarget::Executable {
                        file: psiphon_bin.as_path().into(),
                        args: "",
                    },
                    file: desktop_path().join("Psiphon3.lnk").into(),
                },
            ],
        ),
        Pipeline::new(
            "Install Minecraft (Java Edition)",
            vec![
                Step::DownloadFile {
                    file: minecraft_dir.join("Minecraft.exe").into(),
                    res: RemoteResource::Url("https://launcher.mojang.com/download/Minecraft.exe"),
                },
                Step::CreateShortcut {
                    target: ShortcutTarget::Executable {
                        file: minecraft_dir.join("Minecraft.exe").into(),

                        args: "",
                    },
                    file: desktop_path().join("Minecraft.lnk").into(),
                },
            ],
        ),
        Pipeline::new(
            "Create C:\\ Shortcut",
            vec![Step::CreateShortcut {
                target: ShortcutTarget::Path {
                    path: Path::new("C:\\").into(),
                },
                file: desktop_path().join("OSDisk (C).lnk").into(),
            }],
        ),
        Pipeline::new(
            "Create PowerShell Shortcut",
            vec![Step::CreateShortcut {
                target: ShortcutTarget::Executable {
                    file: Path::new("C:\\WINDOWS\\system32\\cmd.exe").into(),

                    args: "/C powershell",
                },
                file: desktop_path().join("PowerShell.lnk").into(),
            }],
        ),
    ];

    let ctx = Context {
        reqwest: reqwest::Client::builder()
            .build()
            .into_diagnostic()
            .wrap_err("Could not initialize HTTP client.")?,
    };

    for utility in utilities {
        utility.invoke(&ctx).await?;
    }

    Ok(())
}
